@@ -1,9 +1,9 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2004-2006 Edgewall Software
+# Copyright (C) 2004-2009 Edgewall Software
 # Copyright (C) 2004 Daniel Lundin <daniel@edgewall.com>
 # Copyright (C) 2005-2006 Christopher Lenz <cmlenz@gmx.de>
-# Copyright (C) 2006 Christian Boos <cboos@neuf.fr>
+# Copyright (C) 2006-2007 Christian Boos <cboos@neuf.fr>
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -59,15 +59,248 @@ an object that can be `read()`.
 import re
 from StringIO import StringIO
 
+from genshi import Markup, Stream
+from genshi.core import TEXT, START, END, START_NS, END_NS
+from genshi.builder import Fragment, tag
+from genshi.input import HTMLParser
+
 from trac.config import IntOption, ListOption, Option
 from trac.core import *
-from trac.util import sorted
+from trac.resource import Resource
+from trac.util import reversed, sorted, Ranges
 from trac.util.text import to_utf8, to_unicode
-from trac.util.html import escape, Markup, Fragment, html
+from trac.util.translation import _
 
 
 __all__ = ['get_mimetype', 'is_binary', 'detect_unicode', 'Mimeview',
-           'content_to_unicode']
+           'content_to_unicode', 'Context']
+
+class Context(object):
+    """Rendering context.
+
+    A rendering context specifies ''how'' the content should be rendered.
+    It holds together all the needed contextual information that will be
+    needed by individual renderer components.
+
+    To that end, a context keeps track of the Href instance (`.href`) which
+    should be used as a base for building URLs.
+    
+    It also provides a `PermissionCache`  (`.perm`) which can be used to
+    restrict the output so that only the authorized information is shown.
+    
+    A rendering context may also be associated to some Trac resource which
+    will be used as the implicit reference when rendering relative links
+    or for retrieving relative content and can be used to retrieve related
+    metadata.
+
+    Contexts can be nested, and a new context can be created from an existing
+    context using the call syntax. The previous context can be retrieved using
+    the `.parent` attribute.
+    
+    For example, when rendering a wiki text of a wiki page, the context will
+    be associated to a resource identifying that wiki page.
+    
+    If that wiki text contains a `[[TicketQuery]]` wiki macro, the macro will
+    set up nested contexts for each matching ticket that will be used for
+    rendering the ticket descriptions.
+
+    :since: version 0.11
+    """
+
+    def __init__(self, resource, href=None, perm=None):
+        """Directly create a `Context`.
+
+        :param resource: the associated resource
+        :type resource: `Resource`
+        :param href: an `Href` object suitable for creating URLs
+        :param perm: a `PermissionCache` object used for restricting the
+                     generated output to "authorized" informations only.
+
+        The actual `.perm` attribute of the rendering context will be bound
+        to the given `resource` so that fine-grained permission checks will
+        apply to that.
+        """
+        self.parent = None #: The parent context, if any
+        self.resource = resource
+        self.href = href
+        self.perm = resource and perm and perm(resource) or perm
+        self._hints = None
+
+    def from_request(cls, req, resource=None, id=False, version=False,
+                     absurls=False):
+        """Create a rendering context from a request.
+
+        The `perm` and `href` properties of the context will be initialized
+        from the corresponding properties of the request object.
+
+        >>> from trac.test import Mock, MockPerm
+        >>> req = Mock(href=Mock(), perm=MockPerm())
+        >>> context = Context.from_request(req)
+        >>> context.href is req.href
+        True
+        >>> context.perm is req.perm
+        True
+
+        :param req: the HTTP request object
+        :param resource: the `Resource` object or realm
+        :param id: the resource identifier
+        :param version: the resource version
+        :param absurls: whether URLs generated by the ``href`` object should
+                        be absolute (including the protocol scheme and host
+                        name)
+        :return: a new rendering context
+        :rtype: `Context`
+        """
+        if req:
+            href = absurls and req.abs_href or req.href
+            perm = req.perm
+        else:
+            href = None
+            perm = None
+        self = cls(Resource(resource, id=id, version=version), href=href,
+                   perm=perm)
+        self.req = req
+        return self
+    from_request = classmethod(from_request)
+
+    def __repr__(self):
+        path = []
+        context = self
+        while context:
+            if context.resource.realm: # skip toplevel resource 
+                path.append(repr(context.resource))
+            context = context.parent
+        return '<%s %s>' % (type(self).__name__, ' - '.join(reversed(path)))
+
+    def __call__(self, resource=None, id=False, version=False):
+        """Create a nested rendering context.
+
+        `self` will be the parent for the new nested context.
+
+        :param resource: either a `Resource` object or the realm string for a
+                         resource specification to be associated to the new
+                         context. If `None`, the resource will be the same
+                         as the resource of the parent context.
+        :param id: the identifier part of the resource specification
+        :param version: the version of the resource specification
+        :return: the new context object
+        :rtype: `Context`
+
+        >>> context = Context('wiki', 'WikiStart')
+        >>> ticket1 = Resource('ticket', 1)
+        >>> context('ticket', 1).resource == ticket1
+        True
+        >>> context(ticket1).resource is ticket1
+        True
+        >>> context(ticket1)().resource is ticket1
+        True
+        """
+        if resource:
+            resource = Resource(resource, id=id, version=version)
+        else:
+            resource = self.resource
+        context = Context(resource, href=self.href, perm=self.perm)
+        context.parent = self
+
+        # hack for context instances created by from_request()
+        # this is needed because various parts of the code rely on a request
+        # object being available, but that will hopefully improve in the
+        # future
+        if hasattr(self, 'req'):
+            context.req = self.req
+
+        return context
+
+    def __contains__(self, resource):
+        """Check whether a resource is in the rendering path.
+
+        The primary use for this check is to avoid to render the content of a
+        resource if we're already embedded in a context associated to that
+        resource.
+
+        :param resource: a `Resource` specification which will be checked for
+        """
+        context = self
+        while context:
+            if context.resource and \
+                   context.resource.realm == resource.realm and \
+                   context.resource.id == resource.id:
+                # we don't care about version here
+                return True
+            context = context.parent
+
+    # Rendering hints 
+    #
+    # A rendering hint is a key/value pairs that can influence renderers,
+    # wiki formatters and processors in the way they produce their output.
+    # The keys are strings, but the values could be anything.
+    #
+    # In nested contexts, the hints are inherited from their parent context,
+    # unless overriden locally.
+
+    def set_hints(self, **keyvalues):
+        """Set rendering hints for this rendering context.
+
+        >>> ctx = Context('timeline')
+        >>> ctx.set_hints(wiki_flavor='oneliner', shorten_lines=True)
+        >>> t_ctx = ctx('ticket', 1)
+        >>> t_ctx.set_hints(wiki_flavor='html', preserve_newlines=True)
+        >>> (t_ctx.get_hint('wiki_flavor'), t_ctx.get_hint('shorten_lines'), \
+             t_ctx.get_hint('preserve_newlines'))
+        ('html', True, True)
+        >>> (ctx.get_hint('wiki_flavor'), ctx.get_hint('shorten_lines'), \
+             ctx.get_hint('preserve_newlines'))
+        ('oneliner', True, None)
+        """
+        if self._hints is None:
+            self._hints = {}
+            hints = self._parent_hints()
+            if hints is not None:
+                self._hints.update(hints)
+        self._hints.update(keyvalues)
+
+    def get_hint(self, hint, default=None):
+        """Retrieve a rendering hint from this context or an ancestor context.
+
+        >>> ctx = Context('timeline')
+        >>> ctx.set_hints(wiki_flavor='oneliner')
+        >>> t_ctx = ctx('ticket', 1)
+        >>> t_ctx.get_hint('wiki_flavor')
+        'oneliner'
+        >>> t_ctx.get_hint('preserve_newlines', True)
+        True
+        """
+        hints = self._hints
+        if hints is None:
+            hints = self._parent_hints()
+            if hints is None:
+                return default
+        return hints.get(hint, default)
+
+    def has_hint(self, hint):
+        """Test whether a rendering hint is defined in this context or in some
+        ancestor context.
+
+        >>> ctx = Context('timeline')
+        >>> ctx.set_hints(wiki_flavor='oneliner')
+        >>> t_ctx = ctx('ticket', 1)
+        >>> t_ctx.has_hint('wiki_flavor')
+        True
+        >>> t_ctx.has_hint('preserve_newlines')
+        False
+        """
+        hints = self._hints
+        if hints is None:
+            hints = self._parent_hints()
+            if hints is None:
+                return False
+        return hint in hints
+
+    def _parent_hints(self):
+        p = self.parent
+        while p and p._hints is None:
+            p = p.parent
+        return p and p._hints
 
 
 # Some common MIME types and their associated keywords and/or file extensions
@@ -82,6 +315,8 @@ KNOWN_MIME_TYPES = {
     'application/x-yaml':     ['yml', 'yaml'],
     
     'application/rss+xml':    ['rss'],
+    'application/xsl+xml':    ['xsl'],
+    'application/xslt+xml':   ['xslt'],
     
     'image/x-icon':           ['ico'],
     'image/svg+xml':          ['svg'],
@@ -93,18 +328,20 @@ KNOWN_MIME_TYPES = {
     'text/plain':             ['txt', 'TXT', 'text', 'README', 'INSTALL',
                                'AUTHORS', 'COPYING', 'ChangeLog', 'RELEASE'],
     'text/xml':               ['xml'],
-    'text/xsl':               ['xsl'],
     'text/x-csrc':            ['c', 'xs'],
     'text/x-chdr':            ['h'],
     'text/x-c++src':          ['cc', 'CC', 'cpp', 'C'],
     'text/x-c++hdr':          ['hh', 'HH', 'hpp', 'H'],
+    'text/x-csharp':          ['cs'],
     'text/x-diff':            ['diff', 'patch'],
     'text/x-eiffel':          ['e'],
     'text/x-elisp':           ['el'],
     'text/x-fortran':         ['f'],
     'text/x-haskell':         ['hs'],
+    'text/x-ini':             ['ini', 'cfg'],
     'text/x-javascript':      ['js'],
     'text/x-objc':            ['m', 'mm'],
+    'text/x-ocaml':           ['ml', 'mli'],
     'text/x-makefile':        ['make', 'mk',
                                'Makefile', 'makefile', 'GNUMakefile'],
     'text/x-pascal':          ['pas'],
@@ -137,9 +374,10 @@ for t, exts in KNOWN_MIME_TYPES.items():
 
 # Simple builtin autodetection from the content using a regexp
 MODE_RE = re.compile(
-    r"#!(?:[/\w.-_]+/)?(\w+)|"               # look for shebang
+    r"#!.+?env (\w+)|"                       # look for shebang with env
+    r"#!(?:[/\w.-_]+/)?(\w+)|"               # look for regular shebang
     r"-\*-\s*(?:mode:\s*)?([\w+-]+)\s*-\*-|" # look for Emacs' -*- mode -*-
-    r"vim:.*?syntax=(\w+)"                   # look for VIM's syntax=<n>
+    r"vim:.*?(?:syntax|filetype|ft)=(\w+)"   # look for VIM's syntax=<n>
     )
 
 def get_mimetype(filename, content=None, mime_map=MIME_MAP):
@@ -163,10 +401,10 @@ def get_mimetype(filename, content=None, mime_map=MIME_MAP):
         except:
             pass
         if not mimetype and content:
-            match = re.search(MODE_RE, content[:1000])
+            match = re.search(MODE_RE, content[:1000] + content[-1000:])
             if match:
-                mode = match.group(1) or match.group(3) or \
-                    match.group(2).lower()
+                mode = match.group(1) or match.group(2) or match.group(4) or \
+                    match.group(3).lower()
                 if mode in mime_map:
                     # 3) mimetype from the content, using the `MODE_RE`
                     return mime_map[mode]
@@ -229,14 +467,18 @@ class IHTMLPreviewRenderer(Interface):
     # support text content where Trac should expand tabs into spaces
     expand_tabs = False
 
+    # indicate whether the output of this renderer is source code that can
+    # be decorated with annotations
+    returns_source = False
+
     def get_quality_ratio(mimetype):
         """Return the level of support this renderer provides for the `content`
         of the specified MIME type. The return value must be a number between
         0 and 9, where 0 means no support and 9 means "perfect" support.
         """
 
-    def render(req, mimetype, content, filename=None, url=None):
-        """Render an XHTML preview of the raw `content`.
+    def render(context, mimetype, content, filename=None, url=None):
+        """Render an XHTML preview of the raw `content` within a Context.
 
         The `content` might be:
          * a `str` object
@@ -255,21 +497,37 @@ class IHTMLPreviewRenderer(Interface):
         be considered to correspond to lines of text in the original content.
         """
 
+
 class IHTMLPreviewAnnotator(Interface):
     """Extension point interface for components that can annotate an XHTML
     representation of file contents with additional information."""
 
     def get_annotation_type():
-        """Return a (type, label, description) tuple that defines the type of
-        annotation and provides human readable names. The `type` element should
-        be unique to the annotator. The `label` element is used as column
-        heading for the table, while `description` is used as a display name to
-        let the user toggle the appearance of the annotation type.
+        """Return a (type, label, description) tuple
+        that defines the type of annotation and provides human readable names.
+        The `type` element should be unique to the annotator.
+        The `label` element is used as column heading for the table,
+        while `description` is used as a display name to let the user
+        toggle the appearance of the annotation type.
+        """
+        
+    def get_annotation_data(context):
+        """Return some metadata to be used by the `annotate_row` method below.
+
+        This will be called only once, before lines are processed.
+        If this raises an error, that annotator won't be used.
         """
 
-    def annotate_line(number, content):
+    def annotate_row(context, row, number, line, data):
         """Return the XHTML markup for the table cell that contains the
-        annotation data."""
+        annotation data.
+
+        `context` is the context corresponding to the content being annotated,
+        `row` is the tr Element being built, `number` is the line number being
+        processed and `line` is the line's actual content.
+        `annotations` is whatever additional data the `get_annotation_data`
+        method decided to provide.
+        """
 
 
 class IContentConverter(Interface):
@@ -289,10 +547,7 @@ class IContentConverter(Interface):
     def convert_content(context, conversion, content): 
         '''Convert the given `AbstractContent` as specified by `Conversion`.
 
-        The conversion takes place in the given formatting `context`.
-        A `context` provides at least a `req` property.
-        If no other specific context object is available, a
-        `ToplevelContext` can be used to wrap the `req` instance.
+        The conversion takes place in the given rendering context.
         
         Return the converted content, which ''must'' be a `MimeContent` object.
         '''
@@ -338,7 +593,7 @@ class Mimeview(Component):
 
     def __init__(self):
         self._mime_map = None
-        
+
     # Public API
 
     def get_supported_conversions(self, mimetype):
@@ -376,8 +631,8 @@ class Mimeview(Component):
         candidates = list(self.get_supported_conversions(mimetype))
         candidates = [c for c in candidates if key in (c[0], c[4])]
         if not candidates:
-            raise TracError('No available MIME conversions from %s to %s' %
-                            (mimetype, key))
+            raise TracError(_('No available MIME conversions from %(old)s '
+                              'to %(new)s', old=mimetype, new=key))
 
         # First successful conversion wins
         for ck, name, ext, input_mimettype, output_mimetype, quality, \
@@ -386,16 +641,16 @@ class Mimeview(Component):
             if not output:
                 continue
             return (output[0], output[1], ext)
-        raise TracError('No available MIME conversions from %s to %s' %
-                        (mimetype, key))
+        raise TracError(_('No available MIME conversions from %(old)s to '
+                          '%(new)s', old=mimetype, new=key))
 
     def get_annotation_types(self):
         """Generator that returns all available annotation types."""
         for annotator in self.annotators:
             yield annotator.get_annotation_type()
 
-    def render(self, req, mimetype, content, filename=None, url=None,
-               annotations=None):
+    def render(self, context, mimetype, content, filename=None, url=None,
+               annotations=None, force_source=False):
         """Render an XHTML preview of the given `content`.
 
         `content` is the same as an `IHTMLPreviewRenderer.render`'s
@@ -410,6 +665,10 @@ class Mimeview(Component):
         """
         if not content:
             return ''
+        if not isinstance(context, Context):
+            # backwards compatibility: the first argument used to be the
+            # request prior to 0.11
+            context = Context.from_request(context)
 
         # Ensure we have a MIME type for this content
         full_mimetype = mimetype
@@ -433,10 +692,16 @@ class Mimeview(Component):
         # First candidate which renders successfully wins.
         # Also, we don't want to expand tabs more than once.
         expanded_content = None
+        errors = []
         for qr, renderer in candidates:
+            if force_source and not getattr(renderer, 'returns_source', False):
+                continue # skip non-source renderers in force_source mode
             try:
-                self.log.debug('Trying to render HTML preview using %s'
-                               % renderer.__class__.__name__)
+                ann_names = annotations and ', '.join(annotations) or \
+                           'No annotations'
+                self.log.debug('Trying to render HTML preview using %s [%s]'
+                               % (renderer.__class__.__name__, ann_names))
+
                 # check if we need to perform a tab expansion
                 rendered_content = content
                 if getattr(renderer, 'expand_tabs', False):
@@ -445,60 +710,97 @@ class Mimeview(Component):
                                                      full_mimetype)
                         expanded_content = content.expandtabs(self.tab_width)
                     rendered_content = expanded_content
-                result = renderer.render(req, full_mimetype, rendered_content,
-                                         filename, url)
+
+                result = renderer.render(context, full_mimetype,
+                                         rendered_content, filename, url)
                 if not result:
                     continue
-                elif isinstance(result, Fragment):
-                    return result
-                elif isinstance(result, basestring):
-                    return Markup(to_unicode(result))
-                elif annotations:
-                    return Markup(self._annotate(result, annotations))
+
+                if not (force_source or getattr(renderer, 'returns_source',
+                                                False)):
+                    # Direct rendering of content
+                    if isinstance(result, basestring):
+                        if not isinstance(result, unicode):
+                            result = to_unicode(result)
+                        return Markup(to_unicode(result))
+                    elif isinstance(result, Fragment):
+                        return result.generate()
+                    else:
+                        return result
+
+                # Render content as source code
+                if annotations:
+                    m = context.req and context.req.args.get('marks') or None
+                    return self._render_source(context, result, annotations,
+                                               m and Ranges(m))
                 else:
-                    buf = StringIO()
-                    buf.write('<div class="code"><pre>')
-                    for line in result:
-                        buf.write(line + '\n')
-                    buf.write('</pre></div>')
-                    return Markup(buf.getvalue())
+                    if isinstance(result, list):
+                        result = Markup('\n').join(result)
+                    return tag.div(class_='code')(tag.pre(result)).generate()
+
             except Exception, e:
-                self.log.warning('HTML preview using %s failed (%s)'
-                                 % (renderer, e), exc_info=True)
+                self.log.warning('HTML preview using %s failed (%s)' %
+                                 (renderer, e), exc_info=True)
+                errors.append((renderer, e))
+        return errors
 
-    def _annotate(self, lines, annotations):
-        buf = StringIO()
-        buf.write('<table class="code"><thead><tr>')
-        annotators = []
+    def _render_source(self, context, stream, annotations, marks=None):
+        from trac.web.chrome import add_warning
+        annotators, labels, titles = {}, {}, {}
         for annotator in self.annotators:
-            atype, alabel, adesc = annotator.get_annotation_type()
+            atype, alabel, atitle = annotator.get_annotation_type()
             if atype in annotations:
-                buf.write('<th class="%s">%s</th>' % (atype, alabel))
-                annotators.append(annotator)
-        buf.write('<th class="content">&nbsp;</th>')
-        buf.write('</tr></thead><tbody>')
+                labels[atype] = alabel
+                titles[atype] = atitle
+                annotators[atype] = annotator
 
-        space_re = re.compile('(?P<spaces> (?: +))|'
-                              '^(?P<tag><\w+.*?>)?( )')
-        def htmlify(match):
-            m = match.group('spaces')
-            if m:
-                div, mod = divmod(len(m), 2)
-                return div * '&nbsp; ' + mod * '&nbsp;'
-            return (match.group('tag') or '') + '&nbsp;'
+        if isinstance(stream, list):
+            stream = HTMLParser(StringIO('\n'.join(stream)))
+        elif isinstance(stream, unicode):
+            text = stream
+            def linesplitter():
+                for line in text.splitlines(True):
+                    yield TEXT, line, (None, -1, -1)
+            stream = linesplitter()
 
-        num = -1
-        for num, line in enumerate(_html_splitlines(lines)):
-            cells = []
-            for annotator in annotators:
-                cells.append(annotator.annotate_line(num + 1, line))
-            cells.append('<td>%s</td>\n' % space_re.sub(htmlify, line))
-            buf.write('<tr>' + '\n'.join(cells) + '</tr>')
-        else:
-            if num < 0:
-                return ''
-        buf.write('</tbody></table>')
-        return buf.getvalue()
+        annotator_datas = []
+        for a in annotations:
+            annotator = annotators[a]
+            try:
+                data = (annotator, annotator.get_annotation_data(context))
+            except TracError, e:
+                self.log.warning("Can't use annotator '%s': %s" %
+                                 (a, e.message))
+                add_warning(context.req, tag.strong("Can't use ", tag.em(a),
+                                               " annotator:") +
+                                    tag.pre(e.message))
+                data = (None, None)
+            annotator_datas.append(data)
+
+        def _head_row():
+            return tag.tr(
+                [tag.th(labels[a], class_=a, title=titles[a])
+                 for a in annotations] +
+                [tag.th(u'\xa0', class_='content')]
+            )
+
+        def _body_rows():
+            for idx, line in enumerate(_group_lines(stream)):
+                row = tag.tr()
+                if marks and idx + 1 in marks:
+                    row(class_='hilite')
+                for annotator, data in annotator_datas:
+                    if annotator:
+                        annotator.annotate_row(context, row, idx+1, line, data)
+                    else:
+                        row.append(tag.td())
+                row.append(tag.td(line))
+                yield row
+
+        return tag.table(class_='code')(
+            tag.thead(_head_row()),
+            tag.tbody(_body_rows())
+        )
 
     def get_max_preview_size(self):
         """Deprecated: use `max_preview_size` attribute directly."""
@@ -581,20 +883,27 @@ class Mimeview(Component):
                                  "option." % (mapping, option))
         return types
     
-    def preview_to_hdf(self, req, content, length, mimetype, filename,
-                       url=None, annotations=None):
+    def preview_data(self, context, content, length, mimetype, filename,
+                     url=None, annotations=None, force_source=False):
         """Prepares a rendered preview of the given `content`.
 
         Note: `content` will usually be an object with a `read` method.
         """        
+        data = {'raw_href': url, 'size': length,
+                'max_file_size': self.max_preview_size,
+                'max_file_size_reached': False,
+                'rendered': None,
+                }
         if length >= self.max_preview_size:
-            return {'max_file_size_reached': True,
-                    'max_file_size': self.max_preview_size,
-                    'raw_href': url}
+            data['max_file_size_reached'] = True
         else:
-            return {'preview': self.render(req, mimetype, content, filename,
-                                           url, annotations),
-                    'raw_href': url}
+            result = self.render(context, mimetype, content, filename, url,
+                                 annotations, force_source=force_source)
+            if isinstance(result, list):
+                data['errors'] = result
+            else:
+                data['rendered'] = result
+        return data
 
     def send_converted(self, req, in_type, content, selector, filename='file'):
         """Helper method for converting `content` and sending it directly.
@@ -605,44 +914,87 @@ class Mimeview(Component):
                                                          content, selector)
         req.send_response(200)
         req.send_header('Content-Type', output_type)
-        req.send_header('Content-Disposition', 'attachment; filename=%s.%s' % 
+        if filename:
+            req.send_header('Content-Disposition', 'filename=%s.%s' % 
                         (filename, ext))
         req.end_headers()
         req.write(content)
-        raise RequestDone        
-        
+        raise RequestDone
 
-def _html_splitlines(lines):
-    """Tracks open and close tags in lines of HTML text and yields lines that
-    have no tags spanning more than one line."""
-    open_tag_re = re.compile(r'<(\w+)(\s.*?)?[^/]?>')
-    close_tag_re = re.compile(r'</(\w+)>')
-    open_tags = []
-    for line in lines:
-        # Reopen tags still open from the previous line
-        for tag in open_tags:
-            line = tag.group(0) + line
-        open_tags = []
 
-        # Find all tags opened on this line
-        for tag in open_tag_re.finditer(line):
-            open_tags.append(tag)
+def _group_lines(stream):
+    space_re = re.compile('(?P<spaces> (?: +))|^(?P<tag><\w+.*?>)?( )')
+    def pad_spaces(match):
+        m = match.group('spaces')
+        if m:
+            div, mod = divmod(len(m), 2)
+            return div * u'\xa0 ' + mod * u'\xa0'
+        return (match.group('tag') or '') + u'\xa0'
 
-        open_tags.reverse()
+    def _generate():
+        stack = []
+        def _reverse():
+            for event in reversed(stack):
+                if event[0] is START:
+                    yield END, event[1][0], event[2]
+                else:
+                    yield END_NS, event[1][0], event[2]
 
-        # Find all tags closed on this line
-        for ctag in close_tag_re.finditer(line):
-            for otag in open_tags:
-                if otag.group(1) == ctag.group(1):
-                    open_tags.remove(otag)
-                    break
+        for kind, data, pos in stream:
+            if kind is TEXT:
+                lines = data.split('\n')
+                if lines:
+                    # First element
+                    for e in stack:
+                        yield e
+                    yield kind, lines.pop(0), pos
+                    for e in _reverse():
+                        yield e
+                    # Subsequent ones, prefix with \n
+                    for line in lines:
+                        yield TEXT, '\n', pos
+                        for e in stack:
+                            yield e
+                        yield kind, line, pos
+                        for e in _reverse():
+                            yield e
+            else:
+                if kind is START or kind is START_NS:
+                    stack.append((kind, data, pos))
+                elif kind is END or kind is END_NS:
+                    stack.pop()
+                else:
+                    yield kind, data, pos
 
-        # Close all tags still open at the end of line, they'll get reopened at
-        # the beginning of the next line
-        for tag in open_tags:
-            line += '</%s>' % tag.group(1)
+    buf = []
+    
+    # Fix the \n at EOF.
+    if not isinstance(stream, list):
+        stream = list(stream)
+    found_text = False
+    
+    for i in range(len(stream)-1, -1, -1):
+        if stream[i][0] is TEXT:
+            e = stream[i]
+            # One chance to strip a \n
+            if not found_text and e[1].endswith('\n'):
+                stream[i] = (e[0], e[1][:-1], e[2])
+            if len(e[1]):
+                found_text = True
+                break
+    if not found_text:
+        raise StopIteration
 
-        yield line
+    for kind, data, pos in _generate():
+        if kind is TEXT and data == '\n':
+            yield Stream(buf[:])
+            del buf[:]
+        else:
+            if kind is TEXT:
+                data = space_re.sub(pad_spaces, data)
+            buf.append((kind, data, pos))
+    if buf:
+        yield Stream(buf[:])
 
 
 # -- Default annotators
@@ -656,9 +1008,13 @@ class LineNumberAnnotator(Component):
     def get_annotation_type(self):
         return 'lineno', 'Line', 'Line numbers'
 
-    def annotate_line(self, number, content):
-        return '<th id="L%s"><a href="#L%s">%s</a></th>' % (number, number,
-                                                            number)
+    def get_annotation_data(self, context):
+        return None
+
+    def annotate_row(self, context, row, lineno, line, data):
+        row.append(tag.th(id='L%s' % lineno)(
+            tag.a(lineno, href='#L%s' % lineno)
+        ))
 
 
 # -- Default renderers
@@ -670,6 +1026,7 @@ class PlainTextRenderer(Component):
     implements(IHTMLPreviewRenderer)
 
     expand_tabs = True
+    returns_source = True
 
     TREAT_AS_BINARY = [
         'application/pdf',
@@ -682,15 +1039,13 @@ class PlainTextRenderer(Component):
             return 0
         return 1
 
-    def render(self, req, mimetype, content, filename=None, url=None):
+    def render(self, context, mimetype, content, filename=None, url=None):
         if is_binary(content):
-            self.env.log.debug("Binary data; no preview available")
+            self.log.debug("Binary data; no preview available")
             return
 
-        self.env.log.debug("Using default plain text mimeviewer")
-        content = content_to_unicode(self.env, content, mimetype)
-        for line in content.splitlines():
-            yield escape(line)
+        self.log.debug("Using default plain text mimeviewer")
+        return content_to_unicode(self.env, content, mimetype)
 
 
 class ImageRenderer(Component):
@@ -702,10 +1057,10 @@ class ImageRenderer(Component):
             return 8
         return 0
 
-    def render(self, req, mimetype, content, filename=None, url=None):
+    def render(self, context, mimetype, content, filename=None, url=None):
         if url:
-            return html.DIV(html.IMG(src=url,alt=filename),
-                            class_="image-file")
+            return tag.div(tag.img(src=url, alt=filename),
+                           class_='image-file')
 
 
 class WikiTextRenderer(Component):
@@ -717,7 +1072,7 @@ class WikiTextRenderer(Component):
             return 8
         return 0
 
-    def render(self, req, mimetype, content, filename=None, url=None):
-        from trac.wiki import wiki_to_html
-        return wiki_to_html(content_to_unicode(self.env, content, mimetype),
-                            self.env, req)
+    def render(self, context, mimetype, content, filename=None, url=None):
+        from trac.wiki.formatter import format_to_html
+        return format_to_html(self.env, context,
+                              content_to_unicode(self.env, content, mimetype))
