@@ -12,11 +12,11 @@ def do_upgrade(env, ver, cursor):
     try:
         repos = env.get_repository()
         youngest = repos.get_youngest_rev_in_cache(db) or ''
-        # deleting first, for the 0.11dev and 0.10.4dev users
-        cursor.execute("DELETE FROM system WHERE name=%s",
-                       (CACHE_YOUNGEST_REV,))
-        cursor.execute("INSERT INTO system (name, value) VALUES (%s, %s)",
-                       (CACHE_YOUNGEST_REV, youngest))
     except TracError: # no repository available
-        pass
+        youngest = ''
+    # deleting first, for the 0.11dev and 0.10.4dev users
+    cursor.execute("DELETE FROM system WHERE name=%s",
+                   (CACHE_YOUNGEST_REV,))
+    cursor.execute("INSERT INTO system (name, value) VALUES (%s, %s)",
+                   (CACHE_YOUNGEST_REV, youngest))
         
